package dev.fitko.fitconnect.client.sender.strategies;

import com.nimbusds.jose.jwk.RSAKey;
import dev.fitko.fitconnect.api.domain.model.event.authtags.AuthenticationTags;
import dev.fitko.fitconnect.api.domain.model.metadata.ContentStructure;
import dev.fitko.fitconnect.api.domain.model.metadata.Hash;
import dev.fitko.fitconnect.api.domain.model.metadata.Metadata;
import dev.fitko.fitconnect.api.domain.model.metadata.SignatureType;
import dev.fitko.fitconnect.api.domain.model.metadata.attachment.ApiAttachment;
import dev.fitko.fitconnect.api.domain.model.metadata.attachment.Purpose;
import dev.fitko.fitconnect.api.domain.model.metadata.data.Data;
import dev.fitko.fitconnect.api.domain.model.metadata.data.MimeType;
import dev.fitko.fitconnect.api.domain.model.metadata.data.SubmissionSchema;
import dev.fitko.fitconnect.api.domain.model.submission.CreateSubmission;
import dev.fitko.fitconnect.api.domain.model.submission.SentSubmission;
import dev.fitko.fitconnect.api.domain.model.submission.Submission;
import dev.fitko.fitconnect.api.domain.model.submission.SubmissionForPickup;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.services.Sender;
import dev.fitko.fitconnect.client.sender.model.Attachment;
import dev.fitko.fitconnect.client.sender.model.AttachmentPayload;
import dev.fitko.fitconnect.client.sender.model.SendableSubmission;
import dev.fitko.fitconnect.core.util.StopWatch;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.net.URI;
import java.nio.charset.StandardCharsets;
import java.util.List;
import java.util.Objects;
import java.util.UUID;
import java.util.stream.Collectors;

import static dev.fitko.fitconnect.api.config.defaults.SchemaConfig.METADATA_V_1_0_0;
import static dev.fitko.fitconnect.client.util.SubmissionUtil.buildAuthenticationTags;
import static dev.fitko.fitconnect.client.util.SubmissionUtil.buildCreateSubmission;
import static dev.fitko.fitconnect.client.util.SubmissionUtil.buildSentSubmission;
import static dev.fitko.fitconnect.client.util.SubmissionUtil.buildSubmitSubmission;

public class SendNewSubmissionStrategy {

    private static final Logger LOGGER = LoggerFactory.getLogger(SendNewSubmissionStrategy.class);

    private final Sender sender;

    public SendNewSubmissionStrategy(final Sender sender) {
        this.sender = sender;
    }

    public SentSubmission send(final SendableSubmission sendableSubmission) {

        final UUID destinationId = sendableSubmission.getDestinationId();
        final RSAKey encryptionKey = sender.getEncryptionKeyForDestination(destinationId);

        LOGGER.info("Encrypting attachments ...");
        final List<AttachmentPayload> encryptedAttachments = encryptAndHashAttachments(encryptionKey, sendableSubmission.getAttachments());
        final CreateSubmission newSubmission = buildCreateSubmission(sendableSubmission, encryptedAttachments);

        final SubmissionForPickup announcedSubmission = sender.createSubmission(newSubmission);
        final UUID announcedSubmissionId = announcedSubmission.getSubmissionId();

        uploadAttachments(encryptedAttachments, announcedSubmissionId);

        LOGGER.info("Creating metadata");
        final Metadata metadata = buildMetadata(sendableSubmission, encryptedAttachments);

        final ValidationResult validatedMetadata = sender.validateMetadata(metadata);
        if (validatedMetadata.hasError()) {
            LOGGER.error("Metadata does not match schema", validatedMetadata.getError());
            return null;
        }

        LOGGER.info("Encrypting submission data ...");
        final String encryptedData = sender.encryptBytes(encryptionKey, sendableSubmission.getData().getBytes(StandardCharsets.UTF_8));

        LOGGER.info("Encrypting metadata ...");
        final String encryptedMetadata = sender.encryptObject(encryptionKey, metadata);

        final var startTimeSubmissionUpload = StopWatch.start();
        final Submission submission = sender.sendSubmission(buildSubmitSubmission(announcedSubmissionId, encryptedData, encryptedMetadata));
        LOGGER.info("Uploading submission took {}", StopWatch.stopWithFormattedTime(startTimeSubmissionUpload));

        LOGGER.info("SUCCESSFULLY HANDED IN SUBMISSION ! \n");
        final AuthenticationTags authenticationTags = buildAuthenticationTags(encryptedAttachments, encryptedData, encryptedMetadata);
        return buildSentSubmission(authenticationTags, submission);

    }

    private Metadata buildMetadata(final SendableSubmission sendableSubmission, final List<AttachmentPayload> encryptedAttachments) {
        final String hashedData = sender.createHash(sendableSubmission.getData().getBytes(StandardCharsets.UTF_8));

        final String dataMimeType = sendableSubmission.getDataMimeType();
        final Data data = createData(dataMimeType, hashedData, sendableSubmission.getSchemaUri());
        final List<ApiAttachment> attachmentMetadata = encryptedAttachments.stream().map(this::toHashedAttachment).collect(Collectors.toList());

        final var contentStructure = new ContentStructure();
        contentStructure.setAttachments(attachmentMetadata);
        contentStructure.setData(data);

        final var metadata = new Metadata();
        metadata.setSchema(METADATA_V_1_0_0.toString());
        metadata.setContentStructure(contentStructure);

        setOptionalProperties(sendableSubmission, metadata);

        return metadata;
    }

    private void setOptionalProperties(final SendableSubmission sendableSubmission, final Metadata metadata) {
        if (sendableSubmission.getAuthenticationInformation() != null) {
            metadata.setAuthenticationInformation(sendableSubmission.getAuthenticationInformation());
        }
        if (sendableSubmission.getReplyChannel() != null) {
            metadata.setReplyChannel(sendableSubmission.getReplyChannel());
        }
        if (sendableSubmission.getPaymentInformation() != null) {
            metadata.setPaymentInformation(sendableSubmission.getPaymentInformation());
        }
    }

    private Data createData(final String mimeType, final String hashedData, final URI schemaUri) {
        final var hash = new Hash();
        hash.setContent(hashedData);
        hash.setSignatureType(SignatureType.SHA_512);

        final var submissionSchema = new SubmissionSchema();
        submissionSchema.setMimeType(MimeType.fromValue(mimeType));
        submissionSchema.setSchemaUri(schemaUri);

        final var data = new Data();
        data.setSubmissionSchema(submissionSchema);
        data.setHash(hash);
        return data;
    }

    private ApiAttachment toHashedAttachment(final AttachmentPayload attachmentPayload) {
        final var attachment = new ApiAttachment();
        attachment.setAttachmentId(attachmentPayload.getAttachmentId());
        attachment.setPurpose(Purpose.ATTACHMENT);
        attachment.setFilename(attachmentPayload.getFileName());
        attachment.setMimeType(attachmentPayload.getMimeType());
        attachment.setDescription(attachmentPayload.getDescription());

        final var hash = new Hash();
        hash.setContent(attachmentPayload.getHashedData());
        hash.setSignatureType(SignatureType.SHA_512);
        attachment.setHash(hash);
        return attachment;
    }

    private void uploadAttachments(final List<AttachmentPayload> attachmentPayloads, final UUID submissionId) {
        if (attachmentPayloads.isEmpty()) {
            LOGGER.info("No attachments to upload");
        } else {
            LOGGER.info("Uploading {} attachment(s)", attachmentPayloads.size());
            final var startTimeAttachmentUpload = StopWatch.start();
            attachmentPayloads.forEach(a -> sender.uploadAttachment(submissionId, a.getAttachmentId(), a.getEncryptedData()));
            LOGGER.info("Uploading attachments took {}", StopWatch.stopWithFormattedTime(startTimeAttachmentUpload));
        }
    }

    private List<AttachmentPayload> encryptAndHashAttachments(final RSAKey encryptionKey, final List<Attachment> attachments) {
        return attachments.stream()
                .filter(Objects::nonNull)
                .map(attachment -> encryptAndHashAttachment(encryptionKey, attachment))
                .collect(Collectors.toList());
    }

    private AttachmentPayload encryptAndHashAttachment(final RSAKey encryptionKey, final Attachment attachment) {
        final String encryptedAttachment = sender.encryptBytes(encryptionKey, attachment.getDataAsBytes());
        final String hashedBytes = sender.createHash(attachment.getDataAsBytes());
        return AttachmentPayload.builder()
                .attachmentId(UUID.randomUUID())
                .hashedData(hashedBytes)
                .encryptedData(encryptedAttachment)
                .mimeType(attachment.getMimeType())
                .fileName(attachment.getFileName())
                .description(attachment.getDescription())
                .build();
    }
}
